# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import warnings

from ..base import (ConfigurationComponent, AppReadyConfigurationAction,
                    BeforeConfigConfigurationAction)
from ...configuration.utils import TGConfigError, get_partial_dict
from ...support.converters import asbool, aslogger


class SimpleAuthenticationConfigurationComponent(ConfigurationComponent):
    """Provide support for Simple Authentication.

    Simple Authentication is the standard way to handle authentication
    and authorization in TurboGears. Where every request has an associated
    User object (which might be None) and each user can be part of one or more
    Groups each having a set of Permissions.

    The simple auth is based on ``repoze.who`` and by default sets the
    required ``authenticators``, ``identifiers`` and ``metadata providers``
    for a form based login. The user, its groups and permissions are retrieved
    through the ``authmetadata`` object configured by the application.

    For most cases instead of tweaking the simple authentication options
    you probably just want to change the behaviour of ``authmetadata``
    object in your application configuration.

    Provided options:

        * ``auth_backed``: Which is the backend used to authenticate
          the provided credentials (username and password) against a
          store of credentials. Can be one of:

           - ``"authmetadata"``: Which means the authmetadata object of your
             application will be in charge of verifying the credentials
             through an ``authenticate`` method.
           - ``None``: Which means to disable the primary authenticator (only
             authenticators explicitly provided in ``authenticators`` ooption
             will be used). Most of Turbogears will consider authentication
             as disabled.
           - ``"ming"``: Which means to veirfy the credntials against a MongoDB
             database ( *deprecated* ).
           - ``"sqlalchemy"``: Which means to veirfy the credntials against a
             SQLalchemy database ( *deprecated* )

        * ``skip_authentication``: Disable authentication for tests, the
          user will always be authenticated through a ``REMOTE_USER`` environ
          key which will be considered the authenticated user id when set.
        * ``sa_auth.authmetadata``: authmetadata instance to use as an authenticator.
          This is always applied unless ``auth_backend`` is ``None`` or the provided
          object lacks an ``authenticate`` method.
        * ``sa_auth.log_stream``: Provide a custom logger fo authentication.
          by default the ``auth`` logger is used.
        * ``sa_auth.identifiers``: The identifiers to use to recognise a logged user.
          By default ``('default', None)`` leads to an authentication cookie being
          used to recognise logged users.
        * ``sa_auth.form_identifies``: Whenever to add the form plugin to
          the identifiers. By default ``True``. This allows the form plugin
          to intercept requests and identify the user credentisl from the
          data submitted by the login page. If this is
          disabled, the form will be able to act as a challenger and redirect
          the user to the login page, but it won't be able to actually get the
          user_name/password from the submitted form and provide them to the
          authenticator.
        * ``sa_auth.cookie_secret``: Secret to encode the ``auth_tkt`` cookie.
          This is only required when ``('default', None)`` is listed
          in ``sa_auth.identifiers``.
        * ``sa_auth.authenticators``: List of authenticators used to authenticate
          user against the provided credentials. By default ``('default', None)`` and
          ``cookie`` are the only enabled ones. This means that user will be
          authenticated against an user name and password using the configured
          ``auth_backend`` or will be authenticated through the presence of an
          authentication cookie.
        * ``sa_auth.cookie_name``: Name of the cookie used to authenticate the
          user if the cookie identifer and authenticator are enabled (by default they are).
        * ``sa_auth.login_url``: Url where the login form is displayed if
          ``sa_auth.form_plugin`` is enabled. By default ``"/login"``.
        * ``sa_auth.login_handler``: Url that should handle form submitted authentication
          requests if ``sa_auth.form_plugin`` is enabled and it is allowed to identify.
          By default it's ``"/login_handler"``.
        * ``sa_auth.logout_handler``: Url that should handle logout requests
          if ``sa_auth.form_plugin`` is enabled and it is allowed to identify.
          By default it's ``"/logout_handler"``
        * ``sa_auth.post_login_url``: Where to redirect user after a login.
          Only applied if ``sa_auth.form_plugin`` is enabled and allowed to identify.
        * ``sa_auth.post_logout_url`` Where to redirect user after a logout.
          Only applied if ``sa_auth.form_plugin`` is enabled and allowed to identify.
        * ``sa_auth.login_counter_name``: Parameter used by form login to keep track
          of login attempts if `sa_auth.form_plugin`` is enabled. By default ``__logins``.
        * ``sa_auth.form_plugin``: Provide an alternative login/logout implement
          for form based authentication.  This might make all ``form_plugin`` related
          options unusable.
        * ``sa_auth.mdproviders``: Enable some metadata providers. This are used to
          inject additional user details into the current request. By default it's
          disabled and :class:`tg.appwrappers.identity.IdentityApplicationWrapper` is
          used instead.

    """
    id = "sa_auth"
    SUPPORTED_AUTH_BACKENDS = ("ming", "sqlalchemy", "authmetadata")

    def get_defaults(self):
        return {
            'auth_backend': None,
            'skip_authentication': False,
            'sa_auth.log_stream': logging.getLogger('auth'),
            'sa_auth.form_plugin': None
        }

    def get_coercion(self):
        return {
            'skip_authentication': asbool,
            'sa_auth.log_stream': aslogger,
            'sa_auth.form_identifies': asbool
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('Simple Authentication only works on an ApplicationConfigurator')

        from ...appwrappers.identity import IdentityApplicationWrapper
        configurator.register_application_wrapper(IdentityApplicationWrapper, after=True)

    def _configure(self, conf, app):
        if conf['auth_backend'] not in self.SUPPORTED_AUTH_BACKENDS:
            return

        if not conf['skip_authentication'] and 'sa_auth.cookie_secret' not in conf:
            raise TGConfigError("You must provide a value for authentication cookies secret. "
                                "Make sure that you have an 'sa_auth.cookie_secret' config value.")

    def _add_middleware(self, conf, app):
        """
        Configure authentication and authorization.
        """
        # Start with the current configured authentication options.
        # Depending on the auth backend a new auth_args dictionary
        # can replace this one later on.
        auth_backend = conf['auth_backend']
        if auth_backend not in self.SUPPORTED_AUTH_BACKENDS:
            return app

        auth_args = get_partial_dict('sa_auth', conf)

        # Removing keywords not used by repoze.who:
        auth_args.pop('password_encryption_method', None)

        # Removing authmetadata as is not used by repoze.who:
        tgauthmetadata = auth_args.pop('authmetadata', None)

        try:
            pos = auth_args['authenticators'].index(('default', None))
        except KeyError:
            # Didn't specify authenticators, setup default one
            pos = None
        except ValueError:
            # Specified authenticators and default is not in there
            # so we want to skip default TG auth configuration.
            pos = -1

        if pos is None or pos >= 0:
            if getattr(tgauthmetadata, 'authenticate', None) is not None:
                from tg.configuration.auth import create_default_authenticator
                auth_args, tgauth = create_default_authenticator(tgauthmetadata, **auth_args)
                authenticator = ('tgappauth', tgauth)
            elif auth_backend == "sqlalchemy":  # pragma: no cover
                warnings.warn('sqlauth is deprecated, you should add authenticate method '
                              'to your authmetadata instance in app_cfg', DeprecationWarning, 2)
                from tg.configuration.sqla.auth import create_default_authenticator
                auth_args, sqlauth = create_default_authenticator(**auth_args)
                authenticator = ('sqlauth', sqlauth)
            elif auth_backend == "ming":  # pragma: no cover
                warnings.warn('mingauth is deprecated, you should add authenticate method '
                              'to your authmetadata instance in app_cfg', DeprecationWarning, 2)
                from tg.configuration.mongo.auth import create_default_authenticator
                auth_args, mingauth = create_default_authenticator(**auth_args)
                authenticator = ('mingauth', mingauth)
            else:
                authenticator = None

            if authenticator is not None:
                if pos is None:
                    auth_args['authenticators'] = [authenticator]
                else:
                    # We make a copy so that we don't modify the original one.
                    auth_args['authenticators'] = auth_args['authenticators']
                    auth_args['authenticators'][pos] = authenticator

        from tg.configuration.auth import setup_auth
        app = setup_auth(app, skip_authentication=conf['skip_authentication'], **auth_args)
        return app



# -*- coding: utf-8 -*-
from ...support.converters import asbool
from ..base import (ConfigurationComponent,
                    BeforeConfigConfigurationAction,
                    AppReadyConfigurationAction)

from logging import getLogger
log = getLogger(__name__)


class StaticsConfigurationComponent(ConfigurationComponent):
    """Provide support for serving Static Files.

    In production, use Apache or another web server to serve static files.

        - ``serve_static``: Enable / Disable serving static files. **Can be set from .ini file**
        - ``paths.static_files``: Directory where the static files should be served from.
                                  Refer to :class:`.PathsConfigurationComponent` for configuration.
    """
    id = 'statics'

    def get_defaults(self):
        return {
            'serve_static': True
        }

    def get_coercion(self):
        return {
            'serve_static': asbool
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
            AppReadyConfigurationAction(self._add_middleware)
        )

    def _configure(self, conf, app):
        if not conf['paths']['static_files']:
            conf['serve_static'] = False

    def _add_middleware(self, conf, app):
        if not conf['serve_static']:
            return app

        from tg.support.statics import StaticsMiddleware
        return StaticsMiddleware(app, conf['paths']['static_files'])


# -*- coding: utf-8 -*-
import os

from tg.util import Bunch
from ..base import ConfigurationComponent, BeforeConfigConfigurationAction


class PathsConfigurationComponent(ConfigurationComponent):
    """Configure application paths.

    States where the application is contained, where
    to load controllers from, templates from, static files
    and which is the application Python package.

    Options:

        * ``package``: The python package containing the web application.
        * ``paths``: A Dictionary of directories where templates, static files
                     and controllers can be found::

                        {
                            'controllers': 'my/path/to/controlllers',
                            'static_files': 'my/path/to/files',
                            'templates': ['list/of/paths/to/templates']
                        )
        * ``static_files``: An alias to ``paths['static_files']`` for convenience.
    """
    id = "paths"

    DEFAULT_PATHS = {
        'root': None,
        'controllers': None,
        'templates': ['.'],
        'static_files': None
    }

    def get_defaults(self):
        return {
            'paths': Bunch(),
            'package': None
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_paths),
            BeforeConfigConfigurationAction(self._configure_root_modules)
        )

    def _configure_paths(self, conf, app):
        try:
            package = conf['package'].__name__
        except (KeyError, AttributeError):
            package = None

        if package is not None:
            conf['package_name'] = conf['package'].__name__
        else:
            conf['package_name'] = None

        if package is None:
            # if we don't have a specified package (ES: we are in minimal mode)
            # we are unable to get paths from the package itself.
            paths = Bunch()
        else:
            root = os.path.dirname(os.path.abspath(conf['package'].__file__))
            paths = Bunch(
                root=root,
                controllers=os.path.join(root, 'controllers'),
                static_files=os.path.join(root, 'public'),
                templates=[os.path.join(root, 'templates')]
            )

        # If the user defined custom paths, then use them instead of the
        # default ones:
        paths.update(conf['paths'])

        # Ensure all paths are set, load default ones otherwise
        for key, val in self.DEFAULT_PATHS.items():
            paths.setdefault(key, val)

        conf['paths'] = paths

    def _configure_root_modules(self, conf, app):
        root_module_path = conf['paths']['root']
        if not root_module_path:
            return None

        base_controller_path = conf['paths']['controllers']
        controller_path = base_controller_path[len(root_module_path)+1:]
        root_controller_module = '.'.join([conf['package_name']] + controller_path.split(os.sep) + ['root'])
        conf['application_root_module'] = root_controller_module


# -*- coding: utf-8 -*-
import logging
from tg.configuration.utils import coerce_config
from tg.support.converters import asbool, asint, aslist
from ..base import ConfigurationComponent, BeforeConfigConfigurationAction, AppReadyConfigurationAction

log = logging.getLogger(__name__)


class SlowRequestsConfigurationComponent(ConfigurationComponent):
    """Provides slow requests reporting for TurboGears through BackLash.

    This is enabled through the ``trace_slowreqs.enable`` option and
    is only enabled when ``debug=false``.

    All the options available for error reporting are configured
    as ``trace_slowreqs.*`` options in your ``app_cfg`` or ``.ini`` files:

        - ``trace_slowreqs.enable`` -> Enable/Disable slow requests reporting,
          by default it's disabled.
        - ``trace_slowreqs.interval`` -> Report requests slower than this value (default: 25s)
        - ``trace_slowreqs.exclude`` -> List of urls that should be excluded

    Slow requests are reported using *EMail* or *Sentry*, the same
    options available in :class:`.ErrorReportingConfigurationComponent` apply
    with ``trace_slowreqs.`` instead of ``trace_errors.``.

    """
    id = "slow_requests"

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_backlash),
            AppReadyConfigurationAction(self._add_middleware)
        )

    def _configure_backlash(self, conf, app):
        slowreqsware = coerce_config(conf, 'trace_slowreqs.', {'smtp_use_tls': asbool,
                                                               'dump_request_size': asint,
                                                               'dump_request': asbool,
                                                               'dump_local_frames': asbool,
                                                               'dump_local_frames_count': asint,
                                                               'enable': asbool,
                                                               'interval': asint,
                                                               'exclude': aslist})

        slowreqsware.setdefault('error_subject_prefix', 'Slow Request: ')
        slowreqsware.setdefault('error_message', 'A request is taking too much time')

        errorware = conf.get('tg.errorware', {})
        for erroropt in errorware:
            slowreqsware.setdefault(erroropt, errorware[erroropt])
        conf['tg.slowreqs'] = slowreqsware

    def _add_middleware(self, conf, app):
        errorware = conf['tg.slowreqs']
        if errorware.get('enable') and not asbool(conf.get('debug')):
            reporters = []

            if errorware.get('error_email'):
                from backlash.tracing.reporters.mail import EmailReporter
                reporters.append(EmailReporter(**errorware))

            if errorware.get('sentry_dsn'):
                from backlash.tracing.reporters.sentry import SentryReporter
                reporters.append(SentryReporter(**errorware))

            if errorware.get('reporters', []):
                for reporter in errorware['reporters']:
                    reporters.append(reporter)

            try:
                import backlash
            except ImportError:  # pragma: no cover
                log.warning("backlash not installed, slow requests reporting won't be available")
            else:
                return backlash.TraceSlowRequestsMiddleware(
                    app, reporters, interval=errorware.get('interval', 25),
                    exclude_paths=errorware.get('exclude', None),
                    context_injectors=[_turbogears_backlash_context]
                )

        return app


def _turbogears_backlash_context(environ):
    tgl = environ.get('tg.locals')
    return {'request': getattr(tgl, 'request', None)}



# -*- coding: utf-8 -*-
import os

from ...configuration.utils import TGConfigError
from ..base import (ConfigurationComponent,
                    BeforeConfigConfigurationAction)

from logging import getLogger
log = getLogger(__name__)


class I18NConfigurationComponent(ConfigurationComponent):
    """Enable support for internationalization.

    Supported Options:

        * ``i18n.lang``: Default web app language if none was requested or
                         detected from browser.
        * ``i18n.enabled``: Enable support for translations.
                            Disabling i18n will speedup requests and all
                            pages will be served as they were in ``i18n.lang`` value.
        * ``localedir``: Where to find translation catalogs.
                         By default it's project root/i18n

    Refer to :class:`.I18NApplicationWrapper` for additional options
    in supporting i18n.
    """
    id = 'i18n'

    def get_defaults(self):
        return {
            'i18n.lang': None,
            'i18n.enabled': True
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('I18N only works on an ApplicationConfigurator')

        from ...appwrappers.i18n import I18NApplicationWrapper
        configurator.register_application_wrapper(I18NApplicationWrapper, after=True)

    def _configure(self, conf, app):
        if conf['paths']['root']:
            conf['localedir'] = os.path.join(conf['paths']['root'], 'i18n')
        else:
            conf['i18n.enabled'] = False



# -*- coding: utf-8 -*-
from tg.support.converters import asbool

from ...configuration.utils import TGConfigError
from ..base import (ConfigurationComponent,
                    BeforeConfigConfigurationAction, AppReadyConfigurationAction)

from logging import getLogger
log = getLogger(__name__)


class ToscaWidgets2ConfigurationComponent(ConfigurationComponent):
    """Support for ToscaWidgets2 based Widgets, Forms and Resources.

    Options:

        * ``tw2.enabled``: Enable ToscaWidgets2 support.
                           For backward compatibility also ``prefer_toscawidgets2``
                           and ``use_toscawidgets2`` options are supported.
        * ``custom_tw2_config``: Additional TW2 Middleware options to provide.
                                 This should be a dictionary of options.

    Some options from other components influence TW2 behaviour too:

        * ``auto_reload_templates``: Enable autoreloading of Widgets templates.
        * ``debug``: Enable debug mode for TW2 (also injects debug resources).
        * ``default_renderer``: Rendering Engine used to render Widgets templates.
        * ``renderers``: Rendering engines that can be used to render Widgets templates.

    """
    id = 'tw2'

    def get_defaults(self):
        return {
            'tw2.enabled': False,
            'custom_tw2_config': {}
        }

    def get_coercion(self):
        return {
            'tw2.enabled': asbool,
            'prefer_toscawidgets2': asbool,
            'use_toscawidgets2': asbool
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def _configure(self, conf, app):
        prefer_tw2 = conf.get('prefer_toscawidgets2')
        use_tw2 = conf.get('use_toscawidgets2')
        if prefer_tw2 or use_tw2:
            # Backward compatibility with <=2.3 options.
            conf['tw2.enabled'] = True

    def _add_middleware(self, conf, app):
        """Configure the ToscaWidgets2 middleware"""
        if not conf['tw2.enabled']:
            return app

        from tw2.core.middleware import Config, TwMiddleware
        from tg.i18n import ugettext, get_lang

        available_renderers = set(conf.get('renderers', []))
        shared_engines = list(available_renderers & set(Config.preferred_rendering_engines))
        if not shared_engines:
            raise TGConfigError(
                'None of the configured rendering engines %s is supported '
                'by ToscaWidgets2, unable to configure ToscaWidgets.' % available_renderers
            )

        default_renderer = conf.get('default_renderer', None)
        if default_renderer in shared_engines:
            tw2_engines = [default_renderer] + shared_engines
            tw2_default_engine = default_renderer
        else:
            # If preferred rendering engine is not available in TW2, fallback to another one
            tw2_engines = shared_engines
            tw2_default_engine = shared_engines[0]

        default_tw2_config = dict(default_engine=tw2_default_engine,
                                  preferred_rendering_engines=tw2_engines,
                                  translator=ugettext,
                                  get_lang=lambda: get_lang(all=False),
                                  auto_reload_templates=conf.get('auto_reload_templates', False),
                                  controller_prefix='/tw2/controllers/',
                                  res_prefix='/tw2/resources/',
                                  debug=conf['debug'],
                                  rendering_extension_lookup={
                                       'mako': ['mak', 'mako'],
                                       'genshi': ['genshi', 'html'],
                                       'jinja': ['jinja', 'jinja2'],
                                       'kajiki': ['kajiki', 'xhtml', 'xml']
                                  })

        default_tw2_config.update(conf.get('custom_tw2_config', {}))
        app = TwMiddleware(app, **default_tw2_config)
        return app



# -*- coding: utf-8 -*-

from .base import Configurator, ConfigurationComponent
from .base import (BeforeConfigConfigurationAction,
                   ConfigReadyConfigurationAction,
                   AppReadyConfigurationAction,
                   EnvironmentLoadedConfigurationAction)
from .application import ApplicationConfigurator
from .minimal import MinimalApplicationConfigurator
from .fullstack import FullStackApplicationConfigurator



# -*- coding: utf-8 -*-
import os

from ...configuration.utils import TGConfigError
from ..base import (ConfigurationComponent,
                    BeforeConfigConfigurationAction)

from logging import getLogger
log = getLogger(__name__)


class SessionConfigurationComponent(ConfigurationComponent):
    """Provide support for sessions through Beaker.

    Session components adds support for setting up the session
    manager used by ``tg.sessions``.

    Options:

        * ``session.enabled``: Add support for sessions in the application.
                               By default sesions are enabled.
        * ``session.data_dir``: Where to store session files, by default
                                the ``$cache_dir/sessions`` is used.

    Refer to :class:`.SessionApplicationWrapper` for all the supported
    options.

    """
    id = 'session'

    def get_defaults(self):
        return {
            'session.enabled': True
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('Sessions only work on an ApplicationConfigurator')

        from ...appwrappers.session import SessionApplicationWrapper
        configurator.register_application_wrapper(SessionApplicationWrapper, after=True)

    def _configure(self, conf, app):
        if 'cache_dir' in conf:
            conf.setdefault('session.data_dir', os.path.join(conf['cache_dir'], 'sessions'))



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from ..base import ConfigurationComponent, ConfigReadyConfigurationAction, AppReadyConfigurationAction
from ...configuration.utils import TGConfigError
from ...support.middlewares import DBSessionRemoverMiddleware


class SQLAlchemyConfigurationComponent(ConfigurationComponent):
    """Support Relational Databases through SQLAlchemy.

    Configures SQLAlchemy connection to the database,
    the automatic session cleanup at the end of each request
    and support for master/slave databases.

    Support for transaction commit/rollback on request state
    is provided by :class:`.TransactionManagerConfigurationComponent`.

    The configured SQLAlchemy engine is made available as
    ``tg.app_globals.sa_engine``. While the configured SQLAlchemy
    session is made available as ``tg.config['SQLASession']``.

    Options:

        * ``use_sqlalchemy``: Enable SQLAlchemy in your application.
        * ``model``: Configure where models could be found, by default
                    they are expected in ``model`` Python module
                    within your application package.
        * ``sqlalchemy.*``: Options provided to SQLAlchemy connection
                           as expected by :func:`.engine_from_config`.
        * ``sqlalchemy.master.*`` Enable Master/Slave replication support
                                 and configure the master connection.
        * ``sqlalchemy.slaves.slavename.*``: Configure ``slavename`` slave in
                                            Master/Slave support and setup its
                                            connection. Multiple entries with
                                            different ``slavename`` values can
                                            be provided.

    """
    id = "sqlalchemy"

    def get_defaults(self):
        return {
            'use_sqlalchemy': False,
        }

    def get_actions(self):
        return (
            ConfigReadyConfigurationAction(self._setup_sqlalchemy),
            AppReadyConfigurationAction(self._add_middleware)
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('SQLAlchemy Support only works on an ApplicationConfigurator')

    def _setup_sqlalchemy(self, conf, app):
        if not conf['use_sqlalchemy']:
            return
        self.setup_sqlalchemy(conf, app)

    def setup_sqlalchemy(self, conf, app):
        """Setup SQLAlchemy database engine"""
        engine = self.create_sqlalchemy_engine(conf)

        conf['tg.app_globals'].sa_engine = engine

        model = self._get_model(conf)
        sqla_session = model.init_model(engine)
        if sqla_session is not None:
            # If init_model returns a specific session, keep it around
            # as the SQLAlchemy Session.
            conf['SQLASession'] = sqla_session

        if 'DBSession' not in conf:
            # If the user hasn't specified a default session, assume
            # he/she uses the default DBSession in model
            conf['DBSession'] = model.DBSession

    def _add_middleware(self, conf, app):
        if not conf['use_sqlalchemy']:
            return app
        return self.add_middleware(conf, app)

    def add_middleware(self, conf, app):
        """Set up middleware that cleans up the sqlalchemy session.

        The default behavior of TG 2 is to clean up the session on every
        request.  Only override this method if you know what you are doing!

        """
        dbsession = conf.get('SQLASession')
        if dbsession is None:
            dbsession = conf['DBSession']
        return DBSessionRemoverMiddleware(dbsession, app)

    def _get_model(self, conf):
        try:
            package_models = conf['package'].model
        except AttributeError:
            package_models = None

        model = conf.get('model', package_models)
        if model is None:
            raise TGConfigError('SQLAlchemy enabled, but no models provided')

        return model

    @staticmethod
    def create_sqlalchemy_engine(conf):
        from sqlalchemy import engine_from_config
        balanced_master = conf.get('sqlalchemy.master.url')
        if not balanced_master:
            engine = engine_from_config(conf, 'sqlalchemy.')
        else:
            engine = engine_from_config(conf, 'sqlalchemy.master.')
            conf['balanced_engines'] = {'master': engine,
                                        'slaves': {},
                                        'all': {'master': engine}}

            all_engines = conf['balanced_engines']['all']
            slaves = conf['balanced_engines']['slaves']
            for entry in conf.keys():
                if entry.startswith('sqlalchemy.slaves.'):
                    slave_path = entry.split('.')
                    slave_name = slave_path[2]
                    if slave_name == 'master':
                        raise TGConfigError('A slave node cannot be named master')
                    slave_config = '.'.join(slave_path[:3])
                    all_engines[slave_name] = slaves[slave_name] = engine_from_config(conf, slave_config + '.')

            if not conf['balanced_engines']['slaves']:
                raise TGConfigError('When running in balanced mode your must specify at least a slave node')

        return engine



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from tg.support.converters import asbool
from ..base import ConfigurationComponent
from ...configuration.utils import TGConfigError


class TransactionManagerConfigurationComponent(ConfigurationComponent):
    """Support for transaction manager.

    The transaction manager will automatically handle transactions
    bound to it according to the request state.

    Whenever a request succeeds the transaction manger will
    commit all transactions, while in case of a failure it will
    rollback the transactions.

    Refer to :class:`.TransactionApplicationWrapper` for list of
    supported options.
    """
    id = "transaction_manager"

    def get_defaults(self):
        return {
            'tm.enabled': True
        }

    def get_coercion(self):
        return {
            'tm.enabled': asbool
        }

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('Transactions Support only works on an ApplicationConfigurator')

        from ...appwrappers.transaction_manager import TransactionApplicationWrapper
        configurator.register_application_wrapper(TransactionApplicationWrapper, after=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import mimetypes
from ..base import ConfigurationComponent, BeforeConfigConfigurationAction


class MimeTypesConfigurationComponent(ConfigurationComponent):
    """Configure known MimeTypes.

    Mimetypes are used by turbogears to detect expected content
    types based on file extensions. For example it's used by
    :class:`.DispatchConfigurationComponent` request extensions
    to serve the right content based on URL path extension.

    Options:

        * ``mimetype_lookup``: Additional mapping from extensions to
                               mimetypes that should be configured.
    """
    id = "mimetypes"

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_mimetypes),
        )

    def _configure_mimetypes(self, conf, app):
        conf['mimetypes'] = mimetypes.MimeTypes()

        lookup = {'.json': 'application/json',
                  '.js': 'application/javascript'}
        lookup.update(conf.get('mimetype_lookup', {}))

        for key, value in lookup.items():
            conf['mimetypes'].add_type(value, key)





# -*- coding: utf-8 -*-
from __future__ import absolute_import
from ..base import (ConfigurationComponent, BeforeConfigConfigurationAction, ConfigReadyConfigurationAction,
                    AppReadyConfigurationAction)
from ...configuration.utils import get_partial_dict, TGConfigError
from ...support.converters import asbool, asint


class MingConfigurationComponent(ConfigurationComponent):
    """Support MongoDB through the Ming Object Document Mapper.

    Configures the ming connection to MongoDB,
    the automatic session cleanup at the end of each request
    and support for the UnitOfWork flushing/discard based
    on request success or failures.

    The configured Ming session is made available as
    ``tg.config['MingSession']``.

    Options:

        * ``ming.enabled``: Enable Ming in your application
        * ``ming.url``: MongoDB url to connect to.
        * ``ming.db``: Database to use. Can also be specified in ``ming.url``.
        * ``ming.autoflush``: Automatically flush the unit of work as the end
                              of each request, unless there was a failure.
        * ``ming.connection.*``: Options provided to the ming engine to configure
                                 the datastore and the ``MongoClient``. Refer to
                                 :class:`.MongoClient` for available options.

    See :class:`.MingApplicationWrapper` for additional
    configuration options.
    """
    id = "ming"

    def get_defaults(self):
        return {
            'use_ming': False
        }

    def get_coercion(self):
        def mongo_read_pref(value):
            from pymongo.read_preferences import ReadPreference
            return getattr(ReadPreference, value)

        return {
            'use_ming': asbool,  # backward compatibility for ming.enabled
            'ming.enabled': asbool,
            'ming.autoflush': asbool,
            'ming.connection.max_pool_size': asint,
            'ming.connection.network_timeout': asint,
            'ming.connection.tz_aware': asbool,
            'ming.connection.safe': asbool,
            'ming.connection.journal': asbool,
            'ming.connection.wtimeout': asint,
            'ming.connection.fsync': asbool,
            'ming.connection.ssl': asbool,
            'ming.connection.read_preference': mongo_read_pref,
            'ming.connection.auto_ensure_indexes': asbool
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self.configure),
            ConfigReadyConfigurationAction(self.setup),
            AppReadyConfigurationAction(self.add_middleware),
        )

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('Ming Support only works on an ApplicationConfigurator')

        from ...appwrappers.mingflush import MingApplicationWrapper
        configurator.register_application_wrapper(MingApplicationWrapper, after=True)

    def configure(self, conf, app):
        try:
            autoflush_enabled = conf['ming.autoflush']
        except KeyError:
            autoflush_enabled = True

        conf.setdefault('ming.enabled', conf.get('use_ming', False))
        conf['ming.autoflush'] = conf['ming.enabled'] and autoflush_enabled

    def setup(self, conf, app):
        """Setup MongoDB database engine using Ming"""
        if not conf['ming.enabled']:
            return

        datastore = self.create_ming_datastore(conf)
        conf['tg.app_globals'].ming_datastore = datastore

        model = self._get_models(conf)
        ming_session = model.init_model(datastore)
        if ming_session is not None:
            # If init_model returns a specific session, keep it around
            # as the MongoDB Session.
            conf['MingSession'] = ming_session

        if 'DBSession' not in conf:
            # If the user hasn't specified a default session, assume
            # he/she uses the default DBSession in model
            conf['DBSession'] = model.DBSession

    def add_middleware(self, conf, app):
        """Set up the ming middleware for the unit of work"""
        if not conf['ming.enabled']:
            return app

        from tg.support.middlewares import MingSessionRemoverMiddleware
        from ming.odm import ThreadLocalODMSession
        return MingSessionRemoverMiddleware(ThreadLocalODMSession, app)

    def _get_models(self, conf):
        try:
            package_models = conf['package'].model
        except AttributeError:
            package_models = None

        model = conf.get('model', package_models)
        if model is None:
            raise TGConfigError('Ming enabled, but no models provided')

        return model

    @staticmethod
    def create_ming_datastore(conf):
        from ming import create_datastore
        url = conf['ming.url']
        database = conf.get('ming.db', '')
        try:
            connection_options = get_partial_dict('ming.connection', conf)
        except AttributeError:
            connection_options = {}
        if database and url[-1] != '/':
            url += '/'
        ming_url = url + database
        return create_datastore(ming_url, **connection_options)


# -*- coding: utf-8 -*-
import logging

from tg.configuration.utils import coerce_config
from tg.support.converters import asbool, asint
from ..base import ConfigurationComponent, BeforeConfigConfigurationAction, AppReadyConfigurationAction

log = logging.getLogger(__name__)


class ErrorReportingConfigurationComponent(ConfigurationComponent):
    """Provides Error reporting through Backlash on TurboGears.

    This is enabled/disabled through the ``debug`` configuration option.
    Currently EMail and Sentry backlash reporters can be enabled.

    All the options available for error reporting are configured
    as ``trace_errors.*`` options in your ``app_cfg`` or ``.ini`` files.

    The available options for **EMail** reporter are:

        - ``trace_errors.enable`` -> Enable or disable error reporting,
          by default is enabled if backlash is available and ``debug=false``
        - ``trace_errors.smtp_server`` -> SMTP Server to connect to for sending emails
        - ``trace_errors.smtp_port`` -> SMTP port to connect to
        - ``trace_errors.from_address`` -> Address sending the error emails
        - ``trace_errors.error_email`` -> Address the error emails should be sent to.
        - ``trace_errors.smtp_username`` -> Username to authenticate on SMTP server.
        - ``trace_errors.smtp_password`` -> Password to authenticate on SMTP server.
        - ``trace_errors.smtp_use_tls`` -> Whenever to enable or not TLS for SMTP.
        - ``trace_errors.error_subject_prefix`` -> Prefix to append to error emails,
          by default ``WebApp Error:`` is prepended.
        - ``trace_errors.dump_request`` -> Whenever to attach a request dump to the email so that
          all request data is provided.
        - ``trace_errors.dump_request_size`` -> Do not dump request if it's bigger than this value,
          useful for uploaded files. By default 50K.
        - ``trace_errors.dump_local_frames`` -> Enable dumping local variables in case of crashes.
        - ``trace_errors.dump_local_frames_count`` -> Dump up to X frames when dumping local variables.
          The default is 2
        - ``trace_errors.reporters`` -> Add custom reporters to error reporting middleware.

    Available options for **Sentry** reporter are:

        - ``trace_errors.sentry_dsn`` -> Sentry instance where to send the errors.

    """

    id = "error_reporting"

    def get_defaults(self):
        return {
            'debug': False,
            'trace_errors.enable': True
        }

    def get_coercion(self):
        return {
            'debug': asbool
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_backlash),
            AppReadyConfigurationAction(self._add_middleware),
        )

    def _configure_backlash(self, conf, app):
        trace_errors_config = coerce_config(conf, 'trace_errors.', {
            'enable': asbool,
            'smtp_use_tls': asbool,
            'dump_request_size': asint,
            'dump_request': asbool,
            'dump_local_frames': asbool,
            'dump_local_frames_count': asint
        })

        trace_errors_config.setdefault('debug', conf.get('debug', False))
        trace_errors_config.setdefault('error_subject_prefix', 'WebApp Error: ')
        trace_errors_config.setdefault('error_message', 'An internal server error occurred')
        conf['tg.errorware'] = trace_errors_config

    def _add_middleware(self, conf, app):
        errorware = conf['tg.errorware']
        if errorware.get('enable', True) and not asbool(conf.get('debug')):
            reporters = []

            if errorware.get('error_email'):
                from backlash.tracing.reporters.mail import EmailReporter
                reporters.append(EmailReporter(**errorware))

            if errorware.get('sentry_dsn'):
                from backlash.tracing.reporters.sentry import SentryReporter
                reporters.append(SentryReporter(**errorware))

            if errorware.get('reporters', []):
                for reporter in errorware['reporters']:
                    reporters.append(reporter)

            try:
                import backlash
            except ImportError:  # pragma: no cover
                log.warning("backlash not installed, email tracebacks won't be available")
            else:
                return backlash.TraceErrorsMiddleware(
                    app, reporters, context_injectors=[_turbogears_backlash_context]
                )
        return app


def _turbogears_backlash_context(environ):
    tgl = environ.get('tg.locals')
    return {'request': getattr(tgl, 'request', None)}



import os
here = os.path.abspath(os.path.dirname(__file__))
exec(compile(open(os.path.join(here, 'tg', 'release.py')).read(), 'release.py', 'exec'), globals(), locals())

from setuptools import find_packages, setup

import sys
py_version = sys.version_info[:2]

if py_version < (2, 7):
    raise RuntimeError('TurboGears2 requires Python 2.7 or better')

if py_version[0] == 3 and py_version < (3, 4):
    raise RuntimeError('When using Python3 TurboGears2 requires at least Python3.4')

test_requirements = ['nose',
                     'zope.sqlalchemy >= 0.4',
                     'repoze.who',
                     'repoze.who.plugins.sa >= 1.0.1',
                     'Genshi >= 0.5.1',
                     'Mako',
                     'WebTest',
                     'backlash >= 0.0.7',
                     'raven',
                     'formencode>=1.3.0a1',
                     'tw2.forms',
                     'Beaker',
                     'sqlalchemy',
                     'jinja2',
                     'coverage',
                     'ming >= 0.8.0',
                     'Kajiki >= 0.4.4']


if py_version[0] == 2:
    test_requirements.extend(['TurboKid >= 1.0.4',
                              'tgming',
                              'tw.forms'])

install_requires=[
    'WebOb >= 1.8.0, < 1.10.0',
    'crank >= 0.8.0, < 0.9.0',
    'repoze.lru',
    'MarkupSafe'
]

setup(
    name='TurboGears2',
    version=version,
    description=description,
    long_description=long_description,
    classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
    ],
    keywords='turbogears',
    author=author,
    author_email=email,
    url=url,
    license=license,
    packages=find_packages(exclude=('ez_setup', 'examples', 'tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
       # Used by Travis and Coverage due to setup.py nosetests
       # causing a coredump when used with coverage
       'testing':test_requirements,
    },
    test_suite='nose.collector',
    tests_require = test_requirements,
    entry_points='''
    '''
)



# -*- coding: utf-8 -*-
from tg.appwrappers.errorpage import ErrorPageApplicationWrapper
from ..base import ConfigurationComponent, BeforeConfigConfigurationAction


class ErrorPagesConfigurationComponent(ConfigurationComponent):
    """Provides support for custom error pages.

    This will enable the required parts to show a custom error page
    when a common HTTP error happens in the TurboGears application.

    Refer to :class:`.ErrorPageApplicationWrapper` for supported options.

    When simple authentication backend is set to ``None`` (auth disabled)
    status code ``401`` will always be handled by custom error pages
    even if not specified in ``errorpage.status_codes``.

    """
    id = 'error_pages'

    def get_defaults(self):
        return {
            'errorpage.enabled': True,
            'errorpage.status_codes': [403, 404]
        }

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_error_pages),
        )

    def on_bind(self, configurator):
        configurator.register_application_wrapper(ErrorPageApplicationWrapper, after=True)

    def _configure_error_pages(self, conf, app):
        if conf.get('auth_backend') is None and 401 not in conf['errorpage.status_codes']:
            # If there's no auth backend configured which traps 401
            # responses we redirect those responses to a nicely
            # formatted error page
            conf['errorpage.status_codes'] = list(conf['errorpage.status_codes']) + [401]




# -*- coding: utf-8 -*-
import os

from ...support.converters import asbool
from ...configuration.utils import TGConfigError
from ..base import ConfigurationComponent, BeforeConfigConfigurationAction


class CachingConfigurationComponent(ConfigurationComponent):
    """Provide support for caching through Beaker.

    Caching components adds support for setting up the cache
    manager used by all caching features of turbogears:

        * The tg.cache object.
        * Templates caching
        * function results caching by key
        * Controller actions caching.

    Supported Options:

        * ``cache.enabled``: Enabled/Disable caching, by default True.

    Refer to :class:`.CacheApplicationWrapper` for additional options.
    """
    id = 'caching'

    def get_defaults(self):
        return {
            'cache.enabled': True
        }

    def get_coercion(self):
        return {
            'cache.enabled': asbool
        }

    def on_bind(self, configurator):
        from ..application import ApplicationConfigurator
        if not isinstance(configurator, ApplicationConfigurator):
            raise TGConfigError('Caching only works on an ApplicationConfigurator')

        from ...appwrappers.caching import CacheApplicationWrapper
        configurator.register_application_wrapper(CacheApplicationWrapper, after=True)

    def get_actions(self):
        return (
            BeforeConfigConfigurationAction(self._configure_caching),
        )

    def _configure_caching(self, conf, app):
        if 'cache_dir' in conf:
            conf.setdefault('cache.data_dir', os.path.join(conf['cache_dir'], 'cache'))
